#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "btdefines.h"
#include "bthandler.h"
#include "btutils.h"
//PLEASECHECK (Victor): should be included from btutils.h
#include "gotype.h"

#ifdef ENABLE_DEBUG_LOG

#define STATE2STR(state) state2str(state)
#define ERROR2STR(error) error2str(error)

#define UPDATE_STATE(state, newstate) {DEBUG_LOG2("state %s -> %s", STATE2STR(state), STATE2STR(newstate)); state = newstate;}
/* Debug buffer */
#define BT_DBGBUF_LEN 4096
static char dbgbuf[BT_DBGBUF_LEN];

#else //ENABLE_DEBUG_LOG

#define STATE2STR(state) "DEBUG DISABLED"
#define ERROR2STR(error) "DEBUG DISABLED"

#define UPDATE_STATE(state, newstate) {state = newstate;}

#endif //ENABLE_DEBUG_LOG

#define INIT_LISTEN_TIMER(c, t) {BTSetTimer(&c->listentotaltr, t);BTSetTimer(&c->listenretrytr, 0);}
#define INIT_CONNECT_TIMER(c, t) {BTSetTimer(&c->connecttotaltr, t);BTSetTimer(&c->connectretrytr, 0);}

int BTHandlerCompare(cBTConnection con, const char *bdaddr, int chan, int profile, int *res)
{
	int res1, res2;
	
	if (!con || !bdaddr || !res)
		return BT_ERR_ARGUMENT;

	res1 = BTSocketCompare(con->sock, bdaddr, chan, &res2);
	if (res1 != BT_ERR_NONE)
		return res1;

	res1 = (con->profile == profile) ? BT_TRUE : BT_FALSE;

	if (res1 == BT_TRUE && res2 == BT_TRUE)
		*res = BT_TRUE;
	else
		*res = BT_FALSE;

	DEBUG_LOG4("%s, %d, %d: res=%d", bdaddr, chan, profile, *res);
	return BT_ERR_NONE;
}

int BTHandlerCreate(BTConnection *con, const char *bdaddr, int chan, int profile, int flags, const char *pin)
{
	DEBUG_LOG5("addr=[%s], chan=%d, prof=%d, flags=%d pin=[%s]", bdaddr, chan, profile, flags, pin);

	BTConnection c;

	if (!con || chan < 0 || chan > 30 || !bdaddr || !pin)
		return BT_ERR_ARGUMENT;
	if (strlen(bdaddr) > (BT_BDADDR_LEN - 1) || strlen(pin) > (BT_PIN_LEN - 1))
		return BT_ERR_ARGUMENT;
	c = (BTConnection)malloc(sizeof(struct sBTConnection));
	if (!c)
		return BT_ERR_NOMEMORY;
	if (BTSocketCreate(&c->sock, bdaddr, chan) != BT_ERR_NONE) {
		DEBUG_LOG("creating socket failed");
		free(c);
		return BT_ERR_FAILED;
	}
	c->state = BT_STATE_UNINITIALIZED;
	c->ppp.state = BT_PPP_STATE_UNINITIALIZED;
	c->profile = profile;
	c->flags = flags;
	strcpy(c->pin, pin);
	c->sendbuf = NULL;
	c->sendbuflen = 0;
	c->error = BT_ERR_NONE;
	if (flags & BT_FLAG_LISTEN) {
		INIT_LISTEN_TIMER(c, BT_TO_LISTEN_TOTAL);
		UPDATE_STATE(c->state, BT_STATE_START_LISTENING);
	} else if (flags & BT_FLAG_OPEN) {
		INIT_CONNECT_TIMER(c, BT_TO_CONNECT_TOTAL);
		UPDATE_STATE(c->state, BT_STATE_START_CONNECTING);
	} else
		UPDATE_STATE(c->state, BT_STATE_INITIALIZED);
	UPDATE_STATE(c->ppp.state, BT_PPP_STATE_UNINITIALIZED);
	c->ppp.refcount = 0;
	c->ppp.pid = -1;
	c->reconnecting = 0;
	
	*con = c;
	return BT_ERR_NONE;
}

int BTHandlerDestroy(BTConnection con)
{
	DEBUG_LOG1("state: %s", STATE2STR(con->state));

	int res;
	
	if (!con)
		return BT_ERR_ARGUMENT;
	res = BTHandlerClose(con);
	free(con);
	
	return res;
}

int BTHandlerStartApp(BTConnection con, const char *param)
{
	DEBUG_LOG5("profile=%d, param=[%s], state=%s, refcount=%d, flags=0x%x", con->profile, param, STATE2STR(con->state), con->ppp.refcount, con->flags);

	if (con->profile != BT_PROFILE_DUN)
		return BT_ERR_NONE;
	if (!con || !param)
		return BT_ERR_ARGUMENT;
	if (strlen(param) > (BT_PPP_PARAM_LEN - 1))
		return BT_ERR_ARGUMENT;
	if (con->state != BT_STATE_INITIALIZED && con->state != BT_STATE_APP_START && con->state != BT_STATE_APP_READY)
		return BT_ERR_STATE;
	con->ppp.refcount++;
	if(con->ppp.refcount > 1 && con->state != BT_STATE_INITIALIZED)
		return BT_ERR_NONE;
	else if (con->ppp.refcount == 1)
		strcpy(con->ppp.param, param);
	UPDATE_STATE(con->state, BT_STATE_APP_START);
	UPDATE_STATE(con->ppp.state, BT_PPP_STATE_INIT);
	
	return BT_ERR_NONE;
}

int BTHandlerStopApp(BTConnection con)
{
	int pid;
	DEBUG_LOG3("profile=%d, state=%s, refcount=%d", con->profile, STATE2STR(con->state), con->ppp.refcount);

	if (!con)
		return BT_ERR_ARGUMENT;
	if (con->profile != BT_PROFILE_DUN)
		return BT_ERR_NONE;
	if (con->state != BT_STATE_APP_START && con->state != BT_STATE_APP_READY && con->state != BT_STATE_INITIALIZED)
		return BT_ERR_STATE;
	if (con->ppp.refcount)
		con->ppp.refcount--;
	if (con->ppp.refcount || con->state == BT_STATE_INITIALIZED)
		return BT_ERR_NONE;
	if (con->state == BT_STATE_APP_START && (con->ppp.state == BT_PPP_STATE_CHECKTTY || con->ppp.state == BT_PPP_STATE_START || con->ppp.state == BT_PPP_STATE_CHECKREADY) )
		BTSocketReleaseTTY(con->sock);

	pid = BTGetFileInt(BT_PPP_PIDFILE);
	if (pid > 0) {
		DEBUG_LOG1("terminating %d", pid);
		kill(pid, SIGTERM);
	}
	BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
	UPDATE_STATE(con->state, BT_STATE_APP_CLOSING);
	
	return BT_ERR_NONE;
}

int BTHandlerSend(BTConnection con, const char *buf, int len, int *sent)
{
#ifdef ENABLE_DEBUG_LOG
	if (len <= BT_DBGBUF_LEN) {
		if (buf && len)
			strncpy(dbgbuf, buf, len);
		dbgbuf[len - 1] = '\0';
		DEBUG_LOG3("buf=[%s], len=%d, state=%s", dbgbuf, len, STATE2STR(con->state));
	} else {
		DEBUG_LOG("debug buffer is too small!");
	}
#endif //ENABLE_DEBUG_LOG

	char *tmpbuf;
	
	if (!con || !buf || !sent)
		return BT_ERR_ARGUMENT;
	if (con->state == BT_STATE_APP_START || con->state == BT_STATE_APP_READY || con->state == BT_STATE_APP_CLOSING)
		return BT_ERR_STATE;
	if (con->state != BT_STATE_CONNECTED || con->sendbuf) {
		if (!con->sendbuf) {
			con->sendbuf = (char*)malloc(len);
			if (!con->sendbuf)
				return BT_ERR_NOMEMORY;
		} else {
			tmpbuf = (char*)realloc(con->sendbuf, con->sendbuflen + len);
			if (!tmpbuf)
				return BT_ERR_NOMEMORY;
			con->sendbuf = tmpbuf;
		}
		memcpy(con->sendbuf + con->sendbuflen, buf, len);
		DEBUG_LOG2("send buffer grown %d -> %d", con->sendbuflen, con->sendbuflen + len);
		con->sendbuflen += len;
		if (con->state == BT_STATE_START_CONNECTING || con->state == BT_STATE_CONNECTING || con->state == BT_STATE_CONNECTED)
			return BT_ERR_NONE;
		if (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING)
			BTSocketStopListen(con->sock);
		con->reconnecting = 0;
		INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
		UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
	} else {
		if (BTSocketSend(con->sock, buf, len , sent, BT_TO_NONE) != BT_ERR_NONE) {
			BTSocketClose(con->sock);
			DEBUG_LOG1("direct send failed, reconnecting %s", strerror(errno));
			INIT_CONNECT_TIMER(con, BT_TO_RECONNECT_TOTAL);
			UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
			return BT_ERR_CONNECTION;
		}
	}
	return BT_ERR_NONE;
}

int BTHandlerReceive(BTConnection con, char *buf, int len, int *received)
{
	int res;

	if (!con || !buf || !received)
		return BT_ERR_ARGUMENT;
	if (con->state != BT_STATE_CONNECTED)
		return BT_ERR_STATE;
	res = BTSocketReceive(con->sock, buf, len, received, BT_TO_NONE);
	switch (res) {
	case BT_ERR_TIMEOUT:
		*received = 0;
		res = BT_ERR_NONE;
		break;
	case BT_ERR_NONE:
#ifdef ENABLE_DEBUG_LOG
		if (*received <= BT_DBGBUF_LEN) {
			if (*received)
				strncpy(dbgbuf, buf, *received);
			buf[*received] = '\0';
			DEBUG_LOG3("buf=[%s], len=%d, received=%d", buf, len, *received);
		} else {
			DEBUG_LOG("debug buffer is too small!");
		}
#endif //ENABLE_DEBUG_LOG
		res = BT_ERR_NONE;
		break;
	case BT_ERR_FAILED:
		DEBUG_LOG1("receive failed, reconnecting: %s", strerror(errno));
		BTSocketClose(con->sock);
		con->reconnecting = 1;
		INIT_CONNECT_TIMER(con, BT_TO_RECONNECT_TOTAL);
		UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
		*received = 0;
		res = BT_ERR_CONNECTION;
		break;
	default:
		DEBUG_LOG("UNHANDLED ERROR!!!");
		res = BT_ERR_FAILED;
		*received = 0;
		break;
	}
	return res;
}

int BTHandlerRelease(BTConnection con)
{
	DEBUG_LOG1("state=%s", STATE2STR(con->state));

	if (!con)
		return BT_ERR_ARGUMENT;
	if (con->state != BT_STATE_START_CONNECTING && con->state != BT_STATE_CONNECTING && con->state != BT_STATE_CONNECTED && con->state != BT_STATE_REMOVE_CONNECTION)
		return BT_ERR_STATE;
	BTSocketClose(con->sock);
	if (con->sendbuf)
		free(con->sendbuf);
	con->sendbuf = NULL;
	con->sendbuflen = 0;
	if (con->flags & BT_FLAG_LISTEN) {
		DEBUG_LOG("release to listening");
		INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
		UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
	} else {
		DEBUG_LOG("release to initialized");
		UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
	}
	
	return BT_ERR_NONE;
}

int BTHandlerClose(BTConnection con)
{
	int pid;

	DEBUG_LOG1("state=%s", STATE2STR(con->state));

	if (!con)
		return BT_ERR_ARGUMENT;
	if (con->state == BT_STATE_APP_START || con->state == BT_STATE_APP_READY || con->state == BT_STATE_APP_CLOSING) {
		if (con->state != BT_STATE_APP_CLOSING) {
			pid = BTGetFileInt(BT_PPP_PIDFILE);
			if (pid > 0) {
				DEBUG_LOG1("terminating %d", pid);
				kill(pid, SIGTERM);
			}
		}
		BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
		UPDATE_STATE(con->state, BT_STATE_CLOSING);
		UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CLOSING);
	} else if (con->state != BT_STATE_CLOSING) {
		BTSocketClose(con->sock);
		if (con->sendbuf)
			free(con->sendbuf);
		con->sendbuf = NULL;
		con->sendbuflen = 0;
		UPDATE_STATE(con->state, BT_STATE_CLOSING);
		UPDATE_STATE(con->ppp.state, BT_PPP_STATE_UNINITIALIZED);
	}
	return BT_ERR_NONE;
}

int BTHandlerGetState(cBTConnection con, int *state)
{
	if (!con || !state)
		return BT_ERR_ARGUMENT;
//	DEBUG_LOG1("state=%s", STATE2STR(con->state));
	*state = con->state;
	
	return BT_ERR_NONE;
}

int BTHandlerGetError(cBTConnection con, int *error)
{
	if (!con || !error)
		return BT_ERR_ARGUMENT;
	DEBUG_LOG1("error=%s", ERROR2STR(con->error));
	*error = con->error;
	
	return BT_ERR_NONE;
}

int BTHandlerSetFlags(BTConnection con, int flags)
{
	DEBUG_LOG3("flags=0x%X -> 0x%X, state=%s", con->flags, flags, STATE2STR(con->state));

	if (!con)
		return BT_ERR_ARGUMENT;
	if (!(con->flags & BT_FLAG_LISTEN) && (flags & BT_FLAG_LISTEN) && con->state == BT_STATE_INITIALIZED) {
		INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
		UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
	}
	else if (!(con->flags & BT_FLAG_OPEN) && (flags & BT_FLAG_OPEN) && (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING || con->state == BT_STATE_INITIALIZED) && !(con->flags & BT_FLAG_DONOTRECONNECT)) {
		if (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING)
			BTSocketStopListen(con->sock);
		con->reconnecting = 0;
		INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
		UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
	}
	con->flags = flags;

	return BT_ERR_NONE;

}
int BTHandlerGetFlags(cBTConnection con, int *flags)
{
	if (!con || !flags)
		return BT_ERR_ARGUMENT;
	*flags = con->flags;
	
	DEBUG_LOG2("flags=0x%X, state=%s", *flags, STATE2STR(con->state));
	return BT_ERR_NONE;
}

int BTHandlerIsProfileActive(cBTConnection con, int profile, int *res)
{
	if (!con || !profile)
		return BT_ERR_ARGUMENT;

	if (con->profile == profile && (con->state == BT_STATE_CONNECTED || con->state == BT_STATE_APP_READY))
		*res = BT_TRUE;
	else
		*res = BT_FALSE;

	return BT_ERR_NONE;
}

int BTHandlerNeedPair(cBTConnection con, int *res)
{
	if (con->state == BT_STATE_START_LISTENING || con->state == BT_STATE_LISTENING ||
	    con ->state == BT_STATE_START_CONNECTING || con->state == BT_STATE_CONNECTING ||
	    con->state == BT_STATE_APP_START)
		*res = BT_TRUE;
	else
		*res = BT_FALSE;

	return BT_ERR_NONE;
}

int BTHandlerRun(BTConnection con)
{
	int i, res, sent, pid;
	char* argv[9];
	char ttyname[BTS_TTY_NAME_LEN];
	char bdaddrstr[BT_BDADDR_LEN];
	struct stat s;

	if (!con)
		return BT_ERR_ARGUMENT;
	switch (con->state) {
		case BT_STATE_UNINITIALIZED:
			break;
		case BT_STATE_INITIALIZED:
			/* clean-up after pppd finished, sometimes it takes ages to bring it down */
			if (con->profile == BT_PROFILE_DUN)
				if (con->ppp.pid > 0)
					if (waitpid(con->ppp.pid, &res, WNOHANG))
						con->ppp.pid = -1;
			break;
		case BT_STATE_START_LISTENING:
			/* put socket into the listening mode */
			if (!BTTimerExpired(&con->listenretrytr) && !BTTimerExpired(&con->listentotaltr))
				break;
			res = BTSocketListen(con->sock, (con->profile == BT_PROFILE_HF) ? BT_CHANNEL_HF : ( (con->profile == BT_PROFILE_HS) ? BT_CHANNEL_HS : con->sock->chan ) );
			if (res != BT_ERR_NONE) {
				if (BTTimerExpired(&con->listentotaltr)) {
					DEBUG_LOG1("cannot start listening, error=%s", ERROR2STR(res));
					UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
				} else {
					BTSetTimer(&con->listenretrytr, BT_TO_LISTEN_RETRY);
				}
			} else {
				UPDATE_STATE(con->state, BT_STATE_LISTENING);
			}
			break;
		case BT_STATE_LISTENING:
			/* try to accept a new connection */
			res = BTSocketAccept(con->sock, BT_TO_NONE);
			switch (res) {
				case BT_ERR_TIMEOUT:
					break;
				case BT_ERR_NONE:
					UPDATE_STATE(con->state, BT_STATE_CONNECTED);
					break;
				default:
					DEBUG_LOG1("cannot accept, error=%s", ERROR2STR(res));
					BTSocketStopListen(con->sock);
					BTSetTimer(&con->listenretrytr, BT_TO_LISTEN_RETRY);
					UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
					break;
			}
			break;
		case BT_STATE_START_CONNECTING:
			/* put the socket into connecting mode */
			if (!BTTimerExpired(&con->connectretrytr) && !BTTimerExpired(&con->connecttotaltr))
				break;
			/* wait till ACL link is down (the hack needed by some phones) */
			if (!(con->flags & BT_FLAG_MULTICHANNEL)) {
				if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
					if (BTIsAclLinkActive(bdaddrstr)) {
						DEBUG_LOG("ACL link is up, waiting...");
						BTSetTimer(&con->aclremovetr, BT_TO_REMOVEACL);
						UPDATE_STATE(con->state, BT_STATE_REMOVE_CONNECTION);
						break;
					}
				}
			}
			res = BTSocketConnect(con->sock, BT_TO_NONE);
			switch (res) {
				case BT_ERR_NONE:
					UPDATE_STATE(con->state, BT_STATE_CONNECTED);
					break;
				case BT_ERR_TIMEOUT:
					UPDATE_STATE(con->state, BT_STATE_CONNECTING);
					break;
				default:
					DEBUG_LOG1("cannot start connecting, error=%s", ERROR2STR(res));
					if (con->flags & BT_FLAG_DONOTRECONNECT) {
						if (con->flags & BT_FLAG_LISTEN) {
							DEBUG_LOG("reconnect disabled, listening");
							INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
							UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
						} else {
							DEBUG_LOG("reconnect disabled, initialized");
							UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
						}
					}
					if (BTTimerExpired(&con->connecttotaltr)) {
						if (con->flags & BT_FLAG_AUTORECONNECT) {
							DEBUG_LOG("autoreconnect enabled, reconnecting");
							INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
						}
						else if (con->flags & BT_FLAG_LISTEN) {
							DEBUG_LOG("number of retries exceeded, listening");
							INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
							UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
						} else {
							DEBUG_LOG("number of retries exceeded, initialized");
							UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
						}
					} else {
						int modelid = BTGetFileInt(BT_FILENAME_MODELID);
						if (!con->reconnecting &&  modelid != GOTYPE_VALENCIA500 && modelid != GOTYPE_VALENCIA700) {
							DEBUG_LOG("reconnecting using short timeout");
							BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY_1);
						} else {
							DEBUG_LOG("reconnecting using normal timeout");
							BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY);
						}
					}
					break;
			}
			break;
		case BT_STATE_REMOVE_CONNECTION:
			/* wait till ACL link is down (the hack needed by some phones) */
			if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
				if (BTIsAclLinkActive(bdaddrstr) == BT_TRUE) {
					if (!BTTimerExpired(&con->aclremovetr))
						break;
					DEBUG_LOG("destroying ACL link.");
					BTDestroyAclLink(bdaddrstr);
				} else {
					DEBUG_LOG("ACL link is down.");
				}
			}
			BTSetTimer(&con->connectretrytr, 0);
			UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
			break;
		case BT_STATE_CONNECTING:
			/* check if connection is ready, no global timeout here - user may need to enter a PIN code for example */
			res = BTSocketIsConnected(con->sock, BT_TO_NONE);
			switch (res) {
				case BT_ERR_TIMEOUT:
					break;
				case BT_ERR_NONE:
					UPDATE_STATE(con->state, BT_STATE_CONNECTED);
					break;
				default:
					DEBUG_LOG1("connecting failed, error=%s", ERROR2STR(res));
					BTSocketClose(con->sock);
					if (con->flags & BT_FLAG_DONOTRECONNECT) {
						if (con->flags & BT_FLAG_LISTEN) {
							DEBUG_LOG("reconnect disabled, listening");
							INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
							UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
						} else {
							DEBUG_LOG("reconnect disabled, initialized");
							UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
						}
					}
					if (BTTimerExpired(&con->connecttotaltr)) {
						if (con->flags & BT_FLAG_AUTORECONNECT) {
							DEBUG_LOG("autoreconnect enabled, reconnecting");
							INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
							UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
						} else if (con->flags & BT_FLAG_LISTEN) {
							DEBUG_LOG("number of retries exceeded, listening");
							INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
							UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
						} else {
							DEBUG_LOG("number of retries exceeded, initialized");
							UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
						}
					} else {
						int modelid = BTGetFileInt(BT_FILENAME_MODELID);
						if (!con->reconnecting &&  modelid != GOTYPE_VALENCIA500 && modelid != GOTYPE_VALENCIA700) {
							DEBUG_LOG("reconnecting using short timeout");
							BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY_1);
						} else {
							DEBUG_LOG("reconnecting using normal timeout");
							BTSetTimer(&con->connectretrytr, BT_TO_CONNECT_RETRY);
						}
						UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
					}
					break;
			}
			break;
		case BT_STATE_CONNECTED:
			if (!con->sendbuf)
				break;
			res = BTSocketSend(con->sock, con->sendbuf, con->sendbuflen, &sent, BT_TO_NONE);
			if (res == BT_ERR_TIMEOUT)
				break;
			if (res != BT_ERR_NONE) {
				DEBUG_LOG("sending buffer: connection is lost");
				BTSocketClose(con->sock);
				if (BTTimerExpired(&con->connecttotaltr)) {
					free(con->sendbuf);
					con->sendbuf = NULL;
					con->sendbuflen = 0;
					if (con->flags & BT_FLAG_AUTORECONNECT) {
						DEBUG_LOG("autoreconnect enabled, reconnecting");
						con->reconnecting = 1;
						INIT_CONNECT_TIMER(con, BT_TO_CONNECT_TOTAL);
						UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
					}
					else if (con->flags & BT_FLAG_LISTEN) {
						DEBUG_LOG("number of retries exceeded, listening");
						INIT_LISTEN_TIMER(con, BT_TO_LISTEN_TOTAL);
						UPDATE_STATE(con->state, BT_STATE_START_LISTENING);
					} else {
						DEBUG_LOG("number of retries exceeded, initialized");
						UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
					}
				} else {
					BTSetTimer(&con->connectretrytr, 0);
					UPDATE_STATE(con->state, BT_STATE_START_CONNECTING);
				}
				break;
			}
			if (sent < con->sendbuflen) {
				DEBUG_LOG("send buffer partially sent");
				memcpy(con->sendbuf, con->sendbuf + sent, con->sendbuflen - sent);
				con->sendbuflen -= sent;
			} else {
				DEBUG_LOG("send buffer purged");
				free(con->sendbuf);
				con->sendbuf = NULL;
				con->sendbuflen = 0;
			}
			break;
		case BT_STATE_APP_START:
			switch(con->ppp.state){
			case BT_PPP_STATE_INIT:
				pid = BTGetFileInt(BT_PPP_PIDFILE);
				if (pid > 0) {
					DEBUG_LOG1("removing previous instance of %s", BT_PPP_NAME);
					kill(pid, SIGTERM);
					BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
					UPDATE_STATE(con->ppp.state, BT_PPP_STATE_REMOVE);
					break;
				}
				if (!(con->flags & BT_FLAG_MULTICHANNEL)) {
					if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
						if (BTIsAclLinkActive(bdaddrstr)) {
							DEBUG_LOG("ACL link is up, waiting.");
							BTSetTimer(&con->aclremovetr, BT_TO_REMOVEACL);
							UPDATE_STATE(con->state, BT_STATE_APP_REMOVE_CONNECTION);
							break;
						}
					}
				}
				UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CREATETTY);
				//fall trough
			case BT_PPP_STATE_CREATETTY:
				if (BTSocketCreateTTY(con->sock) != BT_ERR_NONE) {
					UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
					break;
				}
				BTSetTimer(&con->ppp.ttytr, BT_TO_TTYCREATE);
				UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CHECKTTY);
				//fall though
			case BT_PPP_STATE_CHECKTTY:
				if (BTSocketIsTTYReady(con->sock, &res) != BT_ERR_NONE) {
					BTSocketReleaseTTY(con->sock);
					UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
					break;
				}				
				if (res == BT_FALSE && !BTTimerExpired(&con->ppp.ttytr))
					break;
				if (res == BT_FALSE) {
					BTSocketReleaseTTY(con->sock);
					UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
					break;
				}
				UPDATE_STATE(con->ppp.state, BT_PPP_STATE_START);
				BTSetTimer(&con->ppp.startretrytr, 0);
				BTSetTimer(&con->ppp.starttotaltr, BT_TO_PPPSTARTTOTAL);
				//fall through
			case BT_PPP_STATE_START:
				/* put the socket into connecting mode */
				if (!BTTimerExpired(&con->ppp.startretrytr))
					break;
				i = 0;
				argv[i++] = BT_PPP_NAME;
				if (BTSocketGetTTYName(con->sock, ttyname) != BT_ERR_NONE) {
					DEBUG_LOG("failed to get TTY name");
					BTSocketReleaseTTY(con->sock);
					UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
					break;
				}
				//PLEASECHECK (Victor): work around, removing the PPP lock file
				unlink("/var/lock/LCK..rfcomm0");
				argv[i++] = ttyname;
				argv[i++] = BT_PPP_BAUDRATE;
				if (strlen(con->ppp.param)) {
					argv[i++] = "call";
					argv[i++] = con->ppp.param;
				}
				argv[i++] = "linkname";
				argv[i++] = BT_PPP_LINKNAME;
				argv[i++] = NULL;
				con->ppp.pid = vfork();
				if (!con->ppp.pid) {
					if (execvp(BT_PPP_NAME, argv) < 0) {
						DEBUG_LOG2("%s failed to start: %s", BT_PPP_NAME, strerror(errno));
						exit(1);
					}
				} else if (con->ppp.pid > 0) {
					DEBUG_LOG2("%s started with pid=%d", BT_PPP_NAME, con->ppp.pid);
					BTSetTimer(&con->ppp.readytr, BT_TO_PPPREADY);
					UPDATE_STATE(con->ppp.state, BT_PPP_STATE_CHECKREADY);
				} else {
					DEBUG_LOG1("failed to start %s", BT_PPP_NAME);
					BTSocketReleaseTTY(con->sock);
					UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
				}
				break;
			case BT_PPP_STATE_CHECKREADY:
				if (BTTimerExpired(&con->ppp.readytr) || waitpid(con->ppp.pid, &res, WNOHANG)) {
					if (!BTTimerExpired(&con->ppp.readytr)) {
						if (WIFEXITED(res)) {
							DEBUG_LOG1("terminated normally, exit status =  %d", WEXITSTATUS(res));
							if (WEXITSTATUS(res) == 7 && !BTTimerExpired(&con->ppp.starttotaltr)) {
								DEBUG_LOG("retrying ppp startup: busy device");
								BTSetTimer(&con->ppp.startretrytr, BT_TO_PPPSTARTRETRY);
								UPDATE_STATE(con->ppp.state, BT_PPP_STATE_START);
								break;
							}
						}
						DEBUG_LOG1("%s failed after start", BT_PPP_NAME);
					} else {
						DEBUG_LOG1("%s timeout expired", BT_PPP_NAME);
					}
					BTSocketReleaseTTY(con->sock);
					pid = BTGetFileInt(BT_PPP_PIDFILE);
					if (pid > 0) {
						DEBUG_LOG1("terminating %d", pid);
						kill(pid, SIGTERM);
					}
					BTSetTimer(&con->ppp.removetr, BT_TO_APPREMOVE);
					UPDATE_STATE(con->state, BT_STATE_APP_CLOSING);
				} else if (stat(BT_PPP_UPFILE, &s) >= 0 ) {
					BTSocketReleaseTTY(con->sock);
					UPDATE_STATE(con->state, BT_STATE_APP_READY);
				}
				break;
			case BT_PPP_STATE_REMOVE:
				pid = BTGetFileInt(BT_PPP_PIDFILE);
				if (pid > 0) {
					if (!waitpid(pid, &res, WNOHANG) && !BTTimerExpired(&con->ppp.removetr))
						break;
					pid = BTGetFileInt(BT_PPP_PIDFILE);
					if (pid > 0) {
						DEBUG_LOG1("sending KILL to %s", BT_PPP_NAME);
						kill(pid, SIGKILL);
						waitpid(pid, &res, WNOHANG);
					}
				}
				UPDATE_STATE(con->ppp.state, BT_PPP_STATE_INIT);
				break;
			default:
				break;
			}
			break;
		case BT_STATE_APP_REMOVE_CONNECTION:
			if (BTSocketGetAddress(con->sock, bdaddrstr) == BT_ERR_NONE) {
				if (BTIsAclLinkActive(bdaddrstr) == BT_TRUE) {
					if (!BTTimerExpired(&con->aclremovetr))
						break;
					DEBUG_LOG("destroying ACL link.");
					BTDestroyAclLink(bdaddrstr);
				} else {
					DEBUG_LOG("ACL link is down.");
				}
			}
			UPDATE_STATE(con->state, BT_STATE_APP_START);
			break;
		case BT_STATE_APP_READY:
			if (waitpid(con->ppp.pid, &res, WNOHANG))
				UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
			break;
		case BT_STATE_APP_CLOSING:
			if (BTTimerExpired(&con->ppp.removetr) || waitpid(con->ppp.pid, &res, WNOHANG)) {
				if (!BTTimerExpired(&con->ppp.removetr))
					con->ppp.pid = -1;
				UPDATE_STATE(con->state, BT_STATE_INITIALIZED);
			}
			break;
		case BT_STATE_CLOSING:
			switch(con->ppp.state){
			case BT_PPP_STATE_CLOSING:
				if (BTTimerExpired(&con->ppp.removetr) || waitpid(con->ppp.pid, &res, WNOHANG)) {
					if (!BTTimerExpired(&con->ppp.removetr))
						con->ppp.pid = -1;
					UPDATE_STATE(con->state, BT_STATE_UNINITIALIZED);
				}
				break;
			case BT_PPP_STATE_UNINITIALIZED:
			default:
				UPDATE_STATE(con->state, BT_STATE_UNINITIALIZED);
				break;
			}
			break;
		default:
			break;
			
	}
	return BT_ERR_NONE;
}
